/* search.c

Functions to search values in sorted tables, and to find roots of various
types of functions (linear, quadratic, cubic, user-defined)

    Copyright (C) 1999, 2000 Conrado Badenas <Conrado.Badenas@uv.es>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "search.h"

#include <math.h>

#include "number.h"
#include "macros.h"
#include "basics.h"

unsigned rootlin(Number a,Number b,Number *x1) /* y=ax+b */
{
	if (a==0.) return 0;
	*x1= -b/a; return 1;
}

unsigned rootquad(Number a,Number b,Number c,Number *x1,Number *x2)
{
	Number q;
	
	if (a==0.) return rootlin(b,c,x1);
	q=b*b-4.*a*c;
	if (q<0.) return 0;
	if (q==0.) {*x1= -0.5*b/a;return 1;}
	q= -0.5*(b+sgn(b)*sqrt(q));*x1=q/a;*x2=c/q;
	return 2;
}

unsigned rootcub(Number a,Number b,Number c,Number d,
	Number *x1,Number *x2,Number *x3)
{
	Number q,q3,r,r2,ot=1./3.;

	if (a==0) return rootquad(b,c,d,x1,x2);
	a=1./a;b*=a;c*=a;d*=a;
	q=(b*b-3.*c)/9.;q3=q*q*q;
	r=(2.*b*b*b-9.*b*c+27.*d)/54.;r2=r*r;
	if (q3>r2)
	{
		Number theta=acos(r/sqrt(q3)),m2rq= -2.*sqrt(q);
		
		*x1=m2rq*cos( theta         *ot)-b*ot;
		*x2=m2rq*cos((theta+2.*M_PI)*ot)-b*ot;
		*x3=m2rq*cos((theta+4.*M_PI)*ot)-b*ot;
		return 3;
	}
	else if (q3==r2)
	{
		Number sqrtq=sqrt(q);
		
		*x1= -2.*sqrtq-b*ot;
		*x2=     sqrtq-b*ot;
		return 2;
	}
	a=pow(sqrt(r2-q3)+fabs(r),ot);
	*x1= -sgn(r)*(a+q/a)-b*ot;
	return 1;
}

Number rootfunc(Number (*f)(Number),Number x1,Number x2,
	Number eabs,Number erel,unsigned maxiter)
{
	unsigned i;
	Number f1,fm,root,dx,xm,fdx;

	if ((f1=(*f)(x1))==0.) return x1;
	if ((fm=(*f)(x2))==0.) return x2;
	if (f1*fm>0.)
		FATAL3("rootfunc(): Root not bracketed in [%g,%g]\n",x1,x2)
	if (f1<0.) {root=x1;dx=x2-x1;}
	else       {root=x2;dx=x1-x2;}
	for (i=0;i<maxiter||maxiter==0;i++)
	{
		fm=(*f)(xm=root+(dx*=0.5));
		fdx=fabs(dx);if (fm<=0.) root=xm;
		if (fdx<eabs || fdx<erel*fabs(root) || fm==0.) return root;
	}
	FATAL2("rootfunc(): Max number of iterations (%u) achieved\n",maxiter)
	return 0.; /* Function never reaches here */
}
